/*
 * Copyright  2005 Sun Microsystems Inc. All rights reserved. Use is subject to license terms.
 *  
 */

package com.sunlabs.archsim.fleet.ships.tokenaccumulator;
import com.sunlabs.archsim.base.simulation.Delay;
import com.sunlabs.archsim.fleet.common.Data;
import com.sunlabs.archsim.fleet.common.OutputFunction;
import com.sunlabs.archsim.gasp.Terminal;
import com.sunlabs.archsim.gasp.Action;
import com.sunlabs.archsim.gasp.terminals.InputTerminal;
import com.sunlabs.archsim.gasp.terminals.OutputTerminal;
import com.sunlabs.archsim.globals.Globals;

class OutputAction extends Action {

	/* interface terminals */
    
	private OutputTerminal dataOut;
	private InputTerminal ackDataOut;
	private InputTerminal tokenOut;
	private OutputTerminal ackTokenOut;

        /* internal terminals */
        
        private InputTerminal intDataIn;
        private OutputTerminal ackIntDataIn;
        
        public OutputAction() {
                this("OutputAction");
        }
        
	public OutputAction(String name){
		super(name);
//		OutputFunction outFunction= new OutFunction();
		// associate done with internal done output
		outputFunctions.put(TokenAccumulatorDefaults.DATA_OUT, new OutFunction());
		outputFunctions.put(TokenAccumulatorDefaults.ACK_INTERNAL_DATA_IN, new NackFunction("Nack Function"));
		outputFunctions.put(TokenAccumulatorDefaults.ACK_TOKEN_OUT, new AckFunction("Done"));
	}
	
        /**
         * Attach an InputTerminal and check whether trying to attach the terminal 
         * twice or a terminal of bad type, e.g. OutputTerminal.
         */
	private InputTerminal attachIn(Terminal t, InputTerminal localTerminal) {
            if (localTerminal == null) {
                    if (t instanceof InputTerminal) {
                            t.attachAction(this);
                            numInputs++;
                            terminals.add(t);            
                            return (InputTerminal)t;
                    } else {
                            // the terminal was of a wrong type
                            director.fatalError("OutputAction, " + getName() + ", " + getID() +  
                                                ": Terminal " + t.getName() + " is of a bad type");
                            return null;
                    }
            } else {
                    // this was the second addition of the same name
                    director.fatalError("OutputAction, " + getName() + ", " + getID() 
					+ ": Terminal " + t.getName() + " cannot be attached twice");
                    return null;
            }
        }

        /**
         * Attach an OutputTerminal and check whether trying to attach the terminal 
         * twice or a terminal of bad type, e.g. InputTerminal.
         */
        private OutputTerminal attachOut(Terminal t, OutputTerminal localTerminal) {
            if (localTerminal == null) {
                    if (t instanceof OutputTerminal) {
                            t.attachAction(this);
                            terminals.add(t);            
                            return (OutputTerminal)t;
                    } else {
                            // the terminal was of a wrong type
                            director.fatalError("OutputAction, " + getName() + ", " + getID() +  
                                                ": Terminal " + t.getName() + " is of a bad type");
                            return null;
                    }
            } else {
                    // this was the second addition of the same name
                    director.fatalError("OutputAction, " + getName() + ", " + getID() 
					+ ": Terminal " + t.getName() + " cannot be attached twice");
                    return null;
            }
        }

	/**
	 * Attach a terminal.
	 * Attach only terminals that are needed for this action, ignore the rest.
	 * Self check ensures that all needed terminals are attached
	 */
	public void attach(Terminal t) {
		// get the bas ename for check
		String termName= t.getBaseName();
                
                if (termName.equals(TokenAccumulatorDefaults.DATA_OUT)) {
                        dataOut= attachOut(t, dataOut); 

                } else if (termName.equals(TokenAccumulatorDefaults.ACK_DATA_OUT)) {
                        ackDataOut= attachIn(t, ackDataOut); 
                
                } else if (termName.equals(TokenAccumulatorDefaults.TOKEN_OUT)) {
                        tokenOut= attachIn(t, tokenOut); 
                
                } else if (termName.equals(TokenAccumulatorDefaults.ACK_TOKEN_OUT)) {
                        ackTokenOut= attachOut(t, ackTokenOut); 
                
                } else if (termName.equals(TokenAccumulatorDefaults.INTERNAL_DATA_IN)) {
                        intDataIn= attachIn(t, intDataIn); 
                
                } else if (termName.equals(TokenAccumulatorDefaults.ACK_INTERNAL_DATA_IN)) {
                        ackIntDataIn= attachOut(t, ackIntDataIn); 
                    
                } else {

			// unknown terminal
			director.error("Terminal " 
					+ t.getName() 
					+ " cannot be attached to OutputAction "
					+ getName() + ", id= " + getID()
					+ "; terminal name not recognized");
		}
	
	} // attach
	
	
	/** self check - are all the terminals attached, run the self check from superclass */
	public boolean selfCheck() {
		boolean OK= true;
		boolean check;

                check= (dataOut != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal dataOut");
		}

                check= (ackTokenOut != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal ackTokenOut");
		}
                
                check= (ackIntDataIn != null);
		OK= OK && check;
		if (!check) {
			director.error("Action " + getAlias() + ", id= " + getID()
					+ ", does not have attached terminal intTokenIn2");
		}
		
		// also run the self check on the superclass
		check= super.selfCheck();
		OK= OK && check;
		
		return OK;
	} // selfCheck

        private class OutFunction extends OutputFunction {

                /** Global settings */
                protected Globals globals= Globals.getInstance();

                /**
                 * Stub for constructor without a specific instance name
                 */ 
                public OutFunction() {
                        this("AckFunction");
                }

                /** 
                 * Call supercalss constructor and set delay.
                 */
                public OutFunction(String name) {
                        super(name);
                        Integer fDelay= globals.intValue(TokenAccumulatorDefaults.ACCUMULATOR_OUT_DELAY);
                        functionDelay= (fDelay != null)
                                        ? new Delay(fDelay.intValue())
                                        : TokenAccumulatorDefaults.ACCUMULATOR_OUT_DELAY_DEF;
                }

                public Data value() {
                        // pass on the data from the internal data in input
                        return (Data)OutputAction.this.intDataIn.getData();

                } // value

        } // OutputFunction
        
        /**
         * Simmilar to AckFunction, but returns an OOB token instead of a valid one.
         */
        private class NackFunction extends OutputFunction {

                /** Global settings */
                protected Globals globals= Globals.getInstance();

                public NackFunction() {
                        this("AckFunction");
                }

                /**
                 * Create the function and set dealy
                 */
                public NackFunction(String name) {
                        super(name);
                        Integer fDelay= globals.intValue(TokenAccumulatorDefaults.ACCUMULATOR_OUT_DELAY);
                        functionDelay= (fDelay != null)
                                        ? new Delay(fDelay.intValue())
                                        : TokenAccumulatorDefaults.ACCUMULATOR_OUT_DELAY_DEF;
                }

                /**
                 * @return  an OOB token.
                 */
                public Data value() {
                        // since the out branch is invoked upon receiving an invalid data
                        // return invalid token
                        return new Data(false);

                } // value

        } // OutputFunction
        
} // class OutputAction
